@@ -1,5 +1,5 @@
 /*
- * Copyright © 2007-2009 Gerd Kohlberger <lowfi@chello.at>
+ * Copyright © 2007-2010 Gerd Kohlberger <gerdko gmail com>
  *
  * This file is part of Mousetweaks.
  *
@@ -23,7 +23,6 @@
 
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
-#include <cspi/spi.h>
 #include <dbus/dbus-glib.h>
 #include <X11/extensions/XTest.h>
 
@@ -36,11 +35,6 @@
 #include "mt-cursor.h"
 #include "mt-main.h"
 #include "mt-listener.h"
-#include "mt-accessible.h"
-
-#define GSM_DBUS_NAME      "org.gnome.SessionManager"
-#define GSM_DBUS_PATH      "/org/gnome/SessionManager"
-#define GSM_DBUS_INTERFACE "org.gnome.SessionManager"
 
 enum {
     PRESS = 0,
@@ -50,20 +44,27 @@ enum {
 };
 
 typedef struct _MtCliArgs {
-    gdouble  delay_time;
+    gdouble  ssc_time;
     gdouble  dwell_time;
     gchar   *mode;
     gint     pos_x;
     gint     pos_y;
     gint     threshold;
-    gboolean delay_click;
-    gboolean dwell_click;
+    gboolean ssc_enabled;
+    gboolean dwell_enabled;
     gboolean shutdown;
     gboolean daemonize;
     gboolean ctw;
     gboolean no_animation;
+    gboolean login;
 } MtCliArgs;
 
+static Display *
+mt_main_display (void)
+{
+    return GDK_DISPLAY ();
+}
+
 static GdkScreen *
 mt_main_current_screen (MtData *mt)
 {
@@ -101,43 +102,38 @@ mt_main_check_mouse_orientation (MtData *mt, guint button)
 
 static void
 mt_main_generate_button_event (MtData *mt,
-			       guint   button,
-			       gint    type,
-			       gulong  delay)
+                               guint   button,
+                               gint    type,
+                               gulong  delay)
 {
+    Display *dpy;
+
+    dpy = mt_main_display ();
     button = mt_main_check_mouse_orientation (mt, button);
 
     gdk_error_trap_push ();
     switch (type) {
-	case PRESS:
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, True, delay);
-	    break;
-	case RELEASE:
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, False, delay);
-	    break;
-	case CLICK:
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, True, CurrentTime);
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, False, delay);
-	    break;
-	case DOUBLE_CLICK:
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, True, CurrentTime);
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, False, delay);
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, True, delay);
-	    XTestFakeButtonEvent (mt->xtst_display,
-				  button, False, delay);
-	    break;
-	default:
-	    g_warning ("Unknown sequence.");
-	    break;
+        case PRESS:
+            XTestFakeButtonEvent (dpy, button, True, delay);
+            break;
+        case RELEASE:
+            XTestFakeButtonEvent (dpy, button, False, delay);
+            break;
+        case CLICK:
+            XTestFakeButtonEvent (dpy, button, True, CurrentTime);
+            XTestFakeButtonEvent (dpy, button, False, delay);
+            break;
+        case DOUBLE_CLICK:
+            XTestFakeButtonEvent (dpy, button, True, CurrentTime);
+            XTestFakeButtonEvent (dpy, button, False, delay);
+            XTestFakeButtonEvent (dpy, button, True, delay);
+            XTestFakeButtonEvent (dpy, button, False, delay);
+            break;
+        default:
+            g_warning ("Unknown button sequence.");
+            break;
     }
-    XFlush (mt->xtst_display);
+    gdk_flush ();
     gdk_error_trap_pop ();
 }
 
@@ -355,92 +351,17 @@ dwell_timer_finished (MtTimer *timer, gpointer data)
     }
 }
 
-static gboolean
-eval_func (Accessible *a, gpointer data)
+static void
+ssc_timer_finished (MtTimer *timer, MtData *mt)
 {
-    gchar *name;
-    gboolean found = FALSE;
-
-    name = Accessible_getName (a);
-    if (name) {
-	found = g_str_equal (name, "Window List");
-	SPI_freeString (name);
-    }
-    return found;
-}
-
-static gboolean
-push_func (Accessible *a, gpointer data)
-{
-    MtData *mt = data;
-    AccessibleRole role;
-
-    role = Accessible_getRole (a);
-    if (role != SPI_ROLE_PANEL && role != SPI_ROLE_EMBEDDED)
-	return FALSE;
-
-    if (!mt_accessible_is_visible (a))
-	return FALSE;
-
-    if (Accessible_isComponent (a))
-	return mt_accessible_in_extents (a, mt->pointer_x, mt->pointer_y);
-
-    return TRUE;
-}
-
-static gboolean
-mt_main_use_move_release (MtData *mt)
-{
-    Accessible *point, *search;
-
-    point = mt_accessible_at_point (mt->pointer_x, mt->pointer_y);
-    if (point) {
-	search = mt_accessible_search (point,
-				       MT_SEARCH_TYPE_BREADTH,
-				       eval_func, push_func, mt);
-	Accessible_unref (point);
-	if (search) {
-	    Accessible_unref (search);
-	    return TRUE;
-	}
-    }
-    return FALSE;
-}
-
-static gboolean
-right_click_timeout (gpointer data)
-{
-    MtData *mt = data;
-
-    mt_main_generate_button_event (mt, 3, CLICK, CurrentTime);
-
-    return FALSE;
+    mt->ssc_finished = TRUE;
 }
 
 static void
-delay_timer_finished (MtTimer *timer, gpointer data)
+mt_main_do_secondary_click (MtData *mt)
 {
-    MtData *mt = data;
-    GdkScreen *screen;
-
-    mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
-
-    if (mt->move_release || mt_main_use_move_release (mt)) {
-	/* release the click outside of the focused object to
-	 * abort any action started by button-press.
-	 */
-	screen = mt_main_current_screen (mt);
-	mt_main_generate_motion_event (screen, 0, 0);
-	mt_main_generate_button_event (mt, 1, RELEASE, CurrentTime);
-	mt_main_generate_motion_event (screen, mt->pointer_x, mt->pointer_y);
-    }
-    else {
-	mt_main_generate_button_event (mt, 1, RELEASE, CurrentTime);
-    }
-    /* wait 100 msec before releasing the button again -
-     * gives apps some time to release active grabs, eg: gnome-panel 'move'
-     */
-    g_timeout_add (100, right_click_timeout, data);
+    mt->ssc_finished = FALSE;
+    mt_main_generate_button_event (mt, 3, CLICK, CurrentTime);
 }
 
 static void
@@ -462,7 +383,6 @@ mt_dwell_click_cancel (MtData *mt)
     dwell_restore_single_click (mt);
 }
 
-/* at-spi listener callbacks */
 static void
 global_motion_event (MtListener *listener,
 		     MtEvent    *event,
@@ -470,10 +390,15 @@ global_motion_event (MtListener *listener,
 {
     MtData *mt = data;
 
-    if (mt_timer_is_running (mt->delay_timer)) {
+    if (mt->ssc_enabled) {
 	if (!below_threshold (mt, event->x, event->y)) {
-	    mt_timer_stop (mt->delay_timer);
 	    mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
+
+	    if (mt_timer_is_running (mt->ssc_timer))
+		mt_timer_stop (mt->ssc_timer);
+
+	    if (mt->ssc_finished)
+		mt->ssc_finished = FALSE;
 	}
     }
 
@@ -507,43 +432,31 @@ global_button_event (MtListener *listener,
 {
     MtData *mt = data;
 
-    if (mt->delay_enabled && event->button == 1) {
-	if (event->type == EV_BUTTON_PRESS) {
+    if (mt->ssc_enabled && event->button == 1) {
+	if (event->type == MT_EVENT_BUTTON_PRESS) {
 	    mt->pointer_x = event->x;
 	    mt->pointer_y = event->y;
-	    mt_timer_start (mt->delay_timer);
+	    mt_timer_start (mt->ssc_timer);
 	}
 	else {
-	    mt_timer_stop (mt->delay_timer);
 	    mt_cursor_manager_restore_all (mt_cursor_manager_get_default ());
+
+	    if (mt->ssc_finished)
+		mt_main_do_secondary_click (mt);
+	    else
+		mt_timer_stop (mt->ssc_timer);
 	}
     }
     /*
      * cancel a dwell-click in progress if a physical button
      * is pressed - useful for mixed use-cases and testing
      */
-    if ((event->type == EV_BUTTON_PRESS && mt_timer_is_running (mt->dwell_timer)) ||
-        (event->type == EV_BUTTON_RELEASE && mt->dwell_drag_started)) {
+    if ((event->type == MT_EVENT_BUTTON_PRESS && mt_timer_is_running (mt->dwell_timer)) ||
+        (event->type == MT_EVENT_BUTTON_RELEASE && mt->dwell_drag_started)) {
 	mt_dwell_click_cancel (mt);
     }
 }
 
-static void
-global_focus_event (MtListener *listener,
-		    gpointer    data)
-{
-    MtData *mt = data;
-    Accessible *accessible;
-
-    if (mt->delay_enabled) {
-	accessible = mt_listener_current_focus (listener);
-	/* TODO: check for more objects and conditions.
-	 * Some links don't have jump actions, eg: text-mails in thunderbird.
-	 */
-	mt->move_release = mt_accessible_supports_action (accessible, "jump");
-    }
-}
-
 static gboolean
 cursor_overlay_time (MtData  *mt,
 		     guchar  *image,
@@ -554,6 +467,7 @@ cursor_overlay_time (MtData  *mt,
 {
     GtkWidget *ctw;
     GdkColor c;
+    GtkStyle *style;
     cairo_surface_t *surface;
     cairo_t *cr;
     gdouble target;
@@ -572,7 +486,8 @@ cursor_overlay_time (MtData  *mt,
     }
 
     ctw = mt_ctw_get_window (mt);
-    c = ctw->style->bg[GTK_STATE_SELECTED];
+    style = gtk_widget_get_style (ctw);
+    c = style->bg[GTK_STATE_SELECTED];
     target = mt_timer_get_target (timer);
 
     cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
@@ -668,10 +583,10 @@ gconf_value_changed (GConfClient *client,
 
     if (g_str_equal (key, OPT_THRESHOLD) && value->type == GCONF_VALUE_INT)
 	mt->threshold = gconf_value_get_int (value);
-    else if (g_str_equal (key, OPT_DELAY) && value->type == GCONF_VALUE_BOOL)
-	mt->delay_enabled = gconf_value_get_bool (value);
-    else if (g_str_equal (key, OPT_DELAY_T) && value->type == GCONF_VALUE_FLOAT)
-	mt_timer_set_target (mt->delay_timer, gconf_value_get_float (value));
+    else if (g_str_equal (key, OPT_SSC) && value->type == GCONF_VALUE_BOOL)
+	mt->ssc_enabled = gconf_value_get_bool (value);
+    else if (g_str_equal (key, OPT_SSC_T) && value->type == GCONF_VALUE_FLOAT)
+	mt_timer_set_target (mt->ssc_timer, gconf_value_get_float (value));
     else if (g_str_equal (key, OPT_DWELL) && value->type == GCONF_VALUE_BOOL) {
 	mt->dwell_enabled = gconf_value_get_bool (value);
 	mt_ctw_update_sensitivity (mt);
@@ -722,15 +637,15 @@ get_gconf_options (MtData *mt)
     gdouble val;
 
     mt->threshold = gconf_client_get_int (mt->client, OPT_THRESHOLD, NULL);
-    mt->delay_enabled = gconf_client_get_bool (mt->client, OPT_DELAY, NULL);
+    mt->ssc_enabled = gconf_client_get_bool (mt->client, OPT_SSC, NULL);
     mt->dwell_enabled = gconf_client_get_bool (mt->client, OPT_DWELL, NULL);
     mt->dwell_show_ctw = gconf_client_get_bool (mt->client, OPT_CTW, NULL);
     mt->dwell_mode = gconf_client_get_int (mt->client, OPT_MODE, NULL);
     mt->style = gconf_client_get_int (mt->client, OPT_STYLE, NULL);
     mt->animate_cursor = gconf_client_get_bool (mt->client, OPT_ANIMATE, NULL);
 
-    val = gconf_client_get_float (mt->client, OPT_DELAY_T, NULL);
-    mt_timer_set_target (mt->delay_timer, val);
+    val = gconf_client_get_float (mt->client, OPT_SSC_T, NULL);
+    mt_timer_set_target (mt->ssc_timer, val);
     val = gconf_client_get_float (mt->client, OPT_DWELL_T, NULL);
     mt_timer_set_target (mt->dwell_timer, val);
 
@@ -747,78 +662,24 @@ get_gconf_options (MtData *mt)
     mt->left_handed = gconf_client_get_bool (mt->client, GNOME_MOUSE_ORIENT, NULL);
 }
 
-static void
-mt_main_request_logout (MtData *mt)
-{
-    DBusGConnection *bus;
-    DBusGProxy *proxy;
-
-    bus = dbus_g_bus_get (DBUS_BUS_SESSION, NULL);
-    if (bus) {
-	proxy = dbus_g_proxy_new_for_name (bus,
-					   GSM_DBUS_NAME,
-					   GSM_DBUS_PATH,
-					   GSM_DBUS_INTERFACE);
-	/*
-	 * Call logout method of session manager:
-	 * mode: 0 = normal, 1 = no confirmation, 2 = force
-	 */
-	dbus_g_proxy_call (proxy, "Logout", NULL,
-			   G_TYPE_UINT, 1, G_TYPE_INVALID,
-			   G_TYPE_INVALID);
-	g_object_unref (proxy);
-    }
-}
-
-static gboolean
-accessibility_enabled (MtData *mt,
-		       gint    spi_status)
-{
-    gint ret;
-
-    if (spi_status != 0) {
-	ret = mt_common_show_dialog
-	    (_("Assistive Technology Support is not Enabled"),
-	     _("Mousetweaks requires assistive technologies to be enabled "
-	       "in your session."
-	       "\n\n"
-	       "To enable support for assistive technologies and restart "
-	       "your session, press \"Enable and Log Out\"."),
-	     MT_MESSAGE_LOGOUT);
-
-	if (ret == GTK_RESPONSE_ACCEPT) {
-	    gconf_client_set_bool (mt->client, GNOME_A11Y_KEY, TRUE, NULL);
-	    mt_main_request_logout (mt);
-	}
-	else {
-	    /* reset the selected option again */
-	    if (gconf_client_get_bool (mt->client, OPT_DELAY, NULL))
-		gconf_client_set_bool (mt->client, OPT_DELAY, FALSE, NULL);
-	    if (gconf_client_get_bool (mt->client, OPT_DWELL, NULL))
-		gconf_client_set_bool (mt->client, OPT_DWELL, FALSE, NULL);
-	}
-	return FALSE;
-    }
-    return TRUE;
-}
-
 static MtData *
 mt_data_init (void)
 {
     MtData *mt;
-    gint ev_base, err_base, maj, min;
+    gint nil;
 
     mt = g_slice_new0 (MtData);
-    mt->xtst_display = XOpenDisplay (NULL);
 
-    if (!XTestQueryExtension (mt->xtst_display,
-			      &ev_base, &err_base, &maj, &min)) {
-	XCloseDisplay (mt->xtst_display);
-	g_slice_free (MtData, mt);
-	g_critical ("No XTest extension found. Aborting..");
-	return NULL;
+    if (!XTestQueryExtension (mt_main_display (), &nil, &nil, &nil, &nil))
+    {
+        g_slice_free (MtData, mt);
+        g_critical ("No XTest extension found. Aborting.");
+        return NULL;
     }
 
+    /* continue sending event requests inspite of other grabs */
+    XTestGrabControl (mt_main_display (), True);
+
     mt->client = gconf_client_get_default ();
     gconf_client_add_dir (mt->client, GNOME_MOUSE_DIR,
 			  GCONF_CLIENT_PRELOAD_NONE, NULL);
@@ -827,10 +688,10 @@ mt_data_init (void)
     g_signal_connect (mt->client, "value_changed",
 		      G_CALLBACK (gconf_value_changed), mt);
 
-    mt->delay_timer = mt_timer_new ();
-    g_signal_connect (mt->delay_timer, "finished",
-		      G_CALLBACK (delay_timer_finished), mt);
-    g_signal_connect (mt->delay_timer, "tick",
+    mt->ssc_timer = mt_timer_new ();
+    g_signal_connect (mt->ssc_timer, "finished",
+		      G_CALLBACK (ssc_timer_finished), mt);
+    g_signal_connect (mt->ssc_timer, "tick",
 		      G_CALLBACK (mt_main_timer_tick), mt);
 
     mt->dwell_timer = mt_timer_new ();
@@ -853,7 +714,7 @@ mt_data_init (void)
 static void
 mt_data_free (MtData *mt)
 {
-    g_object_unref (mt->delay_timer);
+    g_object_unref (mt->ssc_timer);
     g_object_unref (mt->dwell_timer);
     g_object_unref (mt->service);
     g_object_unref (mt->client);
@@ -872,13 +733,13 @@ mt_parse_options (int *argc, char ***argv)
     MtCliArgs ca;
     GOptionContext *context;
     GOptionEntry entries[] = {
-	{"enable-dwell", 0, 0, G_OPTION_ARG_NONE, &ca.dwell_click,
+	{"dwell", 0, 0, G_OPTION_ARG_NONE, &ca.dwell_enabled,
 	    N_("Enable dwell click"), 0},
-	{"enable-secondary", 0, 0, G_OPTION_ARG_NONE, &ca.delay_click,
+	{"ssc", 0, 0, G_OPTION_ARG_NONE, &ca.ssc_enabled,
 	    N_("Enable simulated secondary click"), 0},
 	{"dwell-time", 0, 0, G_OPTION_ARG_DOUBLE, &ca.dwell_time,
 	    N_("Time to wait before a dwell click"), "[0.2-3.0]"},
-	{"secondary-time", 0, 0, G_OPTION_ARG_DOUBLE, &ca.delay_time,
+	{"ssc-time", 0, 0, G_OPTION_ARG_DOUBLE, &ca.ssc_time,
 	    N_("Time to wait before a simulated secondary click"), "[0.5-3.0]"},
 	{"dwell-mode", 'm', 0, G_OPTION_ARG_STRING, &ca.mode,
 	    N_("Set the active dwell mode"), "[window|gesture]"},
@@ -896,21 +757,26 @@ mt_parse_options (int *argc, char ***argv)
 	    N_("Disable cursor animations"), 0},
 	{"daemonize", 0, 0, G_OPTION_ARG_NONE, &ca.daemonize,
 	    N_("Start mousetweaks as a daemon"), 0},
+	{"login", 0, 0, G_OPTION_ARG_NONE, &ca.login,
+	    N_("Start mousetweaks in login mode"), 0},
 	{ NULL }
     };
+
     /* init cli arguments */
-    ca.delay_time   = -1.;
-    ca.dwell_time   = -1.;
-    ca.mode         = NULL;
-    ca.pos_x        = -1;
-    ca.pos_y        = -1;
-    ca.threshold    = -1;
-    ca.delay_click  = FALSE;
-    ca.dwell_click  = FALSE;
-    ca.shutdown     = FALSE;
-    ca.daemonize    = FALSE;
-    ca.ctw          = FALSE;
-    ca.no_animation = FALSE;
+    ca.ssc_time      = -1.;
+    ca.dwell_time    = -1.;
+    ca.mode          = NULL;
+    ca.pos_x         = -1;
+    ca.pos_y         = -1;
+    ca.threshold     = -1;
+    ca.ssc_enabled   = FALSE;
+    ca.dwell_enabled = FALSE;
+    ca.shutdown      = FALSE;
+    ca.daemonize     = FALSE;
+    ca.ctw           = FALSE;
+    ca.no_animation  = FALSE;
+    ca.login         = FALSE;
+
     /* parse */
     context = g_option_context_new (_("- GNOME mouse accessibility daemon"));
     g_option_context_add_main_entries (context, entries, GETTEXT_PACKAGE);
@@ -926,8 +792,6 @@ mt_main (int argc, char **argv, MtCliArgs cli_args)
     MtData *mt;
     MtCursorManager *manager;
     MtListener *listener;
-    gint spi_status;
-    gint spi_leaks = 0;
 
     if (mt_pidfile_create () < 0) {
 	g_warning ("Couldn't create PID file.");
@@ -945,24 +809,18 @@ mt_main (int argc, char **argv, MtCliArgs cli_args)
     if (!mt)
 	goto FINISH;
 
-    spi_status = SPI_init ();
-    if (!accessibility_enabled (mt, spi_status)) {
-	mt_data_free (mt);
-	goto FINISH;
-    }
-
     /* load gconf settings */
     get_gconf_options (mt);
 
     /* override with CLI arguments */
-    if (cli_args.dwell_click)
-	mt->dwell_enabled = cli_args.dwell_click;
-    if (cli_args.delay_click)
-	mt->delay_enabled = cli_args.delay_click;
+    if (cli_args.dwell_enabled)
+	mt->dwell_enabled = cli_args.dwell_enabled;
+    if (cli_args.ssc_enabled)
+	mt->ssc_enabled = cli_args.ssc_enabled;
     if (cli_args.dwell_time >= .1 && cli_args.dwell_time <= 3.)
 	mt_timer_set_target (mt->dwell_timer, cli_args.dwell_time);
-    if (cli_args.delay_time >= .1 && cli_args.delay_time <= 3.)
-	mt_timer_set_target (mt->delay_timer, cli_args.delay_time);
+    if (cli_args.ssc_time >= .1 && cli_args.ssc_time <= 3.)
+	mt_timer_set_target (mt->ssc_timer, cli_args.ssc_time);
     if (cli_args.threshold >= 0 && cli_args.threshold <= 30)
 	mt->threshold = cli_args.threshold;
     if (cli_args.ctw)
@@ -978,7 +836,7 @@ mt_main (int argc, char **argv, MtCliArgs cli_args)
 	g_free (cli_args.mode);
     }
 
-    /* init click-type windoe */
+    /* init click-type window */
     if (!mt_ctw_init (mt, cli_args.pos_x, cli_args.pos_y))
 	goto CLEANUP;
 
@@ -996,8 +854,6 @@ mt_main (int argc, char **argv, MtCliArgs cli_args)
                       G_CALLBACK (global_motion_event), mt);
     g_signal_connect (listener, "button_event",
                       G_CALLBACK (global_button_event), mt);
-    g_signal_connect (listener, "focus_changed",
-                      G_CALLBACK (global_focus_event), mt);
 
     gtk_main ();
 
@@ -1006,14 +862,10 @@ mt_main (int argc, char **argv, MtCliArgs cli_args)
     g_object_unref (listener);
 
 CLEANUP:
-    spi_leaks = SPI_exit ();
     mt_data_free (mt);
+
 FINISH:
     mt_pidfile_remove ();
-
-    if (spi_leaks)
-	g_warning ("AT-SPI reported %i leak%s.",
-		   spi_leaks, spi_leaks != 1 ? "s" : "");
 }
 
 int
@@ -1040,29 +892,20 @@ main (int argc, char **argv)
 
 	return ret < 0 ? 1 : 0;
     }
-
     if ((pid = mt_pidfile_is_running ()) >= 0) {
 	g_print ("Mousetweaks is already running. (PID %u)\n", pid);
 	return 1;
     }
-
     if (cli_args.daemonize) {
 	g_print ("Starting daemon.\n");
 	if ((pid = fork ()) < 0) {
 	    g_error ("fork() failed.");
 	    return 1;
 	}
-	else if (pid) {
-	    /* Parent return */
+	else if (pid)
 	    return 0;
-	}
-	else {
-	    /* Child process */
-	    mt_main (argc, argv, cli_args);
-	}
-    }
-    else {
-	mt_main (argc, argv, cli_args);
     }
+    mt_main (argc, argv, cli_args);
+
     return 0;
 }
